# encoding: utf-8

import os
import oauth
import graph
import utils
# import commands


def authenticate(client_id=None, client_secret=None, tokens=[], token=None):
    if not client_id:
        client_id = os.environ.get('FACEBOOK_INSIGHTS_CLIENT_ID')
    if not client_secret:
        client_secret = os.environ.get('FACEBOOK_INSIGHTS_CLIENT_SECRET')

    secrets = client_id and client_secret
    credentials = token or tokens
    if secrets:
        tokens = oauth.authorize(client_id, client_secret)
    elif credentials:
        pass
    else:
        raise KeyError(utils.dedent("""
            Authentication requires either one or more tokens, 
            or a client_id and client_secret passed to this 
            function or available in the environment as 
            FACEBOOK_INSIGHTS_CLIENT_ID and 
            FACEBOOK_INSIGHTS_CLIENT_SECRET.
            """))
    if token:
        return graph.Page(token)
    else:
        return [graph.Page(token) for token in tokens]



# encoding: utf-8

import click

@click.group()
@click.option('--client-id', 
    help='the app id to your facebook app', 
    envvar='FACEBOOK_INSIGHTS_CLIENT_ID')
@click.option('--client-secret', 
    help='the client secret to your facebook app', 
    envvar='FACEBOOK_INSIGHTS_CLIENT_SECRET')
@click.argument('page') # help='the name of the Facebook Page you would like to fetch insights from'
def cli(page=None):
    # TODO: if no page is specified, list all pages that we've saved
    pass

@cli.command()
@click.option('--force', help='ignore any existing credentials')
def authorize():
    pass

@cli.command()
def ls():
    pass

@cli.command()
@cli.option('-s', '--since')
@cli.option('-u', '--until')
@cli.option('-m', '--months', default=0)
@cli.option('-d', '--days', default=0)
@cli.option('-p', '--period')
@cli.option('-M', '--metrics')
def page():
    pass

@cli.command()
@cli.option('--since')
@cli.option('--until')
@cli.option('--months', default=0)
@cli.option('--days', default=0)
@cli.option('--metrics')
def posts():
    # a period doesn't make sense for posts, as all post metrics 
    # are lifetime metrics
    pass

"""
insights posts Guardian\ US --metrics page_impressions --since 2014-05-01 --days 5
"""


# encoding: utf-8

import api
import date
import functional
import server
import url

import re
import textwrap


def dedent(string):
    return textwrap.dedent(string).replace('\n', '')

def extract_quotes(string):
    QUOTE_PATTERN = ur'[\"\u201c\u201e\u00ab]\s?(.+?)\s?[\"\u201c\u201d\u00bb]'
    return re.findall(QUOTE_PATTERN, string)


# encoding: utf-8

def single_serve(message=None, port=5000):
    import logging
    from flask import Flask, Response, request

    app = Flask(__name__)

    log = logging.getLogger('werkzeug')
    log.setLevel(logging.ERROR)

    captured = {}

    @app.route('/')
    def main():
        request.environ.get('werkzeug.server.shutdown')()
        captured.update(dict(request.args.items()))
        if message:
            print message
        return Response(message, mimetype='text/plain')

    app.run(port=port)
    return captured


# encoding: utf-8

import requests
import urlparse

def resolve(url):
    response = requests.head(url, allow_redirects=True)
    return response.url

def base(url):
    base = urlparse.urlsplit(url)[:3]
    url = urlparse.urlunsplit(base + ('', ''))
    return url


# encoding: utf-8

import time
import datetime as builtin_datetime
import pytz
from dateutil import parser
from dateutil.relativedelta import relativedelta

UTC = COMMON_ERA = builtin_datetime.datetime(1, 1, 1, tzinfo=pytz.utc)

def parse(datestring, utc=False):
    if utc:
        return parser.parse(datestring, default=UTC)
    else:
        return parser.parse(datestring)

def timestamp(date, utc=False):
    if isinstance(date, basestring):
        date = parse(date, utc=utc)

    return int(time.mktime(date.timetuple()))

def datetime(obj, utc=False):
    if obj is None:
        return None
    elif isinstance(obj, builtin_datetime.date):
        return obj
    elif isinstance(obj, basestring):
        return parse(obj, utc=utc)
    else:
        raise ValueError("Can only convert strings into dates, received {}".format(obj.__class__))

def date(obj, utc=False):
    # convert datetime into date if needed
    _date = datetime(obj, utc=utc)
    if hasattr(_date, 'date'):
        return _date.date()
    else:
        return _date

formats = {
    'date': lambda date: date, 
    'iso': lambda date: date.isoformat(), 
    'timestamp': lambda date: timestamp(date, utc=True), 
}

def range(start, stop=None, months=0, days=0, format='date'):
    start = date(start, utc=True)
    stop = date(stop, utc=True)

    if days or months:
        if start:
            stop = start + relativedelta(days=days-1, months=months)
        elif stop:
            start = stop - relativedelta(days=days-1, months=months)
        else:
            raise ValueError("Cannot construct a date range from months or days \
                without specifying start or stop.")
    else:
        stop = stop or start

    if format in formats:
        format = formats[format]
    else:
        raise ValueError("{} is an unrecognized date format. Choose from: {}".format(
            format, ", ".join(formats)))

    return format(start), format(stop)



import functools


class memoize:
  def __init__(self, function):
    self.function = function
    self.memoized = {}

  def __call__(self, *args):
    try:
        return self.memoized[args]
    except KeyError:
        self.memoized[args] = self.function(*args)
        return self.memoized[args]


def immutable(method):
    @functools.wraps(method)
    def wrapped_method(self, *vargs, **kwargs):
        obj = self.clone()
        method(obj, *vargs, **kwargs)
        return obj
    return wrapped_method


def noop(value):
    return value


def soak(*vargs, **kwargs):
    pass



# encoding: utf-8

import os
import rauth
import utils
from utils.api import GraphAPI
import urlparse
import webbrowser

"""
To get a longer-lived page access token, 
exchange the User access token for a long-lived one, 
as above, and then request the Page token. The resulting 
page access token will not have any expiry time.
"""


PORT = 5000
REDIRECT_URI = 'http://localhost:{}/'.format(PORT)


class OAuth2Service(rauth.OAuth2Service):
    def __init__(self, *vargs, **kwargs):
        options = dict(
            authorize_url='https://graph.facebook.com/oauth/authorize', 
            access_token_url='https://graph.facebook.com/oauth/access_token', 
            base_url='https://graph.facebook.com/', 
            )
        options.update(**kwargs)
        self.redirect_uri = options.get('redirect_uri')
        if 'redirect_uri' in options:
            del options['redirect_uri']
        super(OAuth2Service, self).__init__(*vargs, **options)

    def get_authorize_url(self, *vargs, **kwargs):
        options = dict(
            scope='manage_pages,read_insights',
            response_type='code',            
            redirect_uri=self.redirect_uri, 
            )
        options.update(**kwargs)
        return super(OAuth2Service, self).get_authorize_url(*vargs, **options)

    def get_access_token(self, code, *vargs, **kwargs):
        long_term = kwargs.get('long_term', False)
        if 'long_term' in kwargs:
            del kwargs['long_term']

        data = dict(
            code=code,
            grant_type='authorization_code', 
            redirect_uri=self.redirect_uri, 
            )
        data.update(kwargs.get('data', {}))
        kwargs['data'] = data
        token = super(OAuth2Service, self).get_access_token(
            *vargs, **kwargs)

        if long_term:
            token = self.get_long_term_token(token)

        return token

    def get_long_term_token(self, short_token):
        graph = GraphAPI()
        data = graph.get('oauth/access_token',   
            grant_type='fb_exchange_token', 
            client_id=self.client_id, 
            client_secret=self.client_secret, 
            fb_exchange_token=short_token, 
            )
        token = dict(urlparse.parse_qsl(data))
        return token['access_token']

    def get_page_tokens(self, long_token):
        graph = GraphAPI(long_token)
        accounts = graph.get('me/accounts')['data']
        return [account['access_token'] for account in accounts]


def authorize_user(facebook, long_term=False):
    authorize_url = facebook.get_authorize_url()
    webbrowser.open(authorize_url)
    qs = utils.server.single_serve(
        message='Authentication flow completed. You may close the browser tab.', 
        port=PORT, 
        )
    return facebook.get_access_token(qs['code'], long_term=long_term)

def authorize_pages(facebook, user_token):
    return facebook.get_page_tokens(user_token)

def authorize(client_id, client_secret):
    facebook = OAuth2Service(
        client_id=client_id, 
        client_secret=client_secret,
        redirect_uri=REDIRECT_URI, 
        )
    user_token = authorize_user(facebook, long_term=True)
    page_tokens = authorize_pages(facebook, user_token)
    return page_tokens



# encoding: utf-8

import copy
import urllib
import facepy


def getdata(obj, key, default=None):
    if key in obj:
        return obj[key]['data']
    else:
        return default


class GraphAPI(facepy.GraphAPI):
    def __init__(self, *vargs, **kwargs):
        self.base = []
        super(GraphAPI, self).__init__(*vargs, **kwargs)

    def _segmentize_endpoint(self, endpoint):
        if not isinstance(endpoint, list):
            endpoint = [endpoint]
        return endpoint

    def _resolve_endpoint(self, endpoint, options={}):
        endpoint = self._segmentize_endpoint(endpoint)
        url = "/".join(self.base + endpoint)

        # remove facepy options, retain everything 
        # that needs to end up in the querystring
        blacklist = ['path', 'page', 'retry', 'data', 'method', 'relative_url']
        for key in options.keys():
            if key in blacklist:
                del options[key]

        if options:
            qs = urllib.urlencode(options)
            return url + '?' + qs
        else:
            return url

    def partial(self, base):
        client = GraphAPI(self.oauth_token)
        client.base = client.base + self._segmentize_endpoint(base)
        return client

    def all(self, endpoint, paramsets, method='GET', body=False, **options):
        """ A nicer interface for batch requests to the 
        same endpoint but with different parameters, e.g. 
        different date ranges. """

        requests = []
        for params in paramsets:
            params = copy.copy(params)
            params.update(options)
            segments = self._segmentize_endpoint(endpoint)
            relative_url = params.get('relative_url')
            url = self._resolve_endpoint(segments + [relative_url], params)
            request = {
                'method': method, 
                'relative_url': url, 
                }

            if body:
                request['body'] = body

            requests.append(request)

        return self.batch(requests)

    def get(self, relative_endpoint=[], *vargs, **kwargs):
        """ An endpoint can be specified as a string
         or as a list of path segments. """

        endpoint = self._resolve_endpoint(relative_endpoint)
        return super(GraphAPI, self).get(endpoint, *vargs, **kwargs)



import facebookinsights as fi
import unittest


class TestBase(unittest.TestCase):
    def setUp(self):
        pages = fi.authorize()
        if not len(pages):
            raise Exception("Cannot proceed with unit testing: \
                the authorized Facebook account does not have access \
                to any Facebook Page Insights.")
        
        self.page = pages[0]


class TestAuthorization(TestBase):
    pass


class TestQuerying(TestBase):
    def test_range_days(self):
        """ It should support various ways of defining date ranges, 
        and these will result in the correct start and end dates. """

    def test_range_months(self):
        """ It should support various ways of defining date ranges, 
        and these will result in the correct start and end dates. """

    def test_page_query(self):
        """ It should be able to run a page query and return a report. """

    def test_post_query(self):
        """ It should be able to run a post query and return a report. """

    def test_query_immutable(self):
        """ It should always refine queries by creating a new query and 
        never modify the original base query. """

    def test_period(self):
        """ It should have shortcut functions that make it easier to
        define the period (day, week, 28_days, lifetime) at which 
        the API should return results. """

    def test_latest(self):
        """ It can limit the total amount of results. """


if __name__ == '__main__':
    unittest.main()


# encoding: utf-8

import os
import re
import math
import copy
import urlparse
import functools
from collections import namedtuple
import pytz
from datetime import datetime
import utils
from utils.api import GraphAPI
from utils.functional import immutable, memoize


class Selection(object):
    def __init__(self, edge):
        self.edge = edge
        self.graph = edge.graph
        self.meta = {
            'since': utils.date.COMMON_ERA, 
            'until': datetime.now(), 
        }
        self.params = {
            'page': False, 
        }

    def clone(self):
        selection = self.__class__(self.edge)
        selection.meta = copy.copy(self.meta)
        selection.params = copy.copy(self.params)
        return selection

    @immutable
    def range(self, since=None, until=None, months=0, days=0):
        if not (since or until or days or months):
            raise ValueError()

        if not until:
            until = datetime.today()

        since, until = utils.date.range(since, until, months, days)
        self.meta['since'] = since
        self.meta['until'] = until
        self.params['page'] = True
        self.params['since'] = utils.date.timestamp(since, utc=True)
        self.params['until'] = utils.date.timestamp(until, utc=True)
        return self

    @immutable
    def since(self, date):
        return self.range(date)

    def __getitem__(self, key):
        return self.get()[key]

    def __iter__(self):
        if not hasattr(self, '_results'):
            self._results = self.get()
        return self._results.__iter__()


class PostSelection(Selection):
    @immutable
    def latest(self, n=1):
        self.params['limit'] = n
        return self

    def find(self, q):
        return self.graph.find(q, 'post')

    def get(self):
        pages = self.graph.get('posts', **self.params)
        if not self.params['page']:
            pages = [pages]

        posts = []
        for page in pages:
            for post in page['data']:
                post = Post(self.edge, post)

                # For date ranges, we can't rely on pagination 
                # because `since` and `until` parameters serve 
                # both as paginators and as range delimiters, 
                # so there will always be a next page.
                if post.created_time.date() >= self.meta['since']:
                    posts.append(post)
                else:
                    return posts

        return posts


class InsightsSelection(Selection):
    @immutable
    def daily(self, metrics=None):
        self.params['period'] = 'day'
        return self._metrics(metrics)

    @immutable
    def weekly(self, metrics=None):
        self.params['period'] = 'week'
        return self._metrics(metrics)

    @immutable
    def monthly(self, metrics=None):
        self.params['period'] = 'days_28'
        return self._metrics(metrics)

    @immutable
    def lifetime(self, metrics=None):
        self.params['period'] = 'lifetime'
        return self._metrics(metrics)

    def _metrics(self, ids=None):
        if ids:
            if isinstance(ids, list):
                self.meta['single'] = False
            else:
                self.meta['single'] = True
                ids = [ids]
            self.meta['metrics'] = ids
        return self

    @property
    def _has_daterange(self):
        return 'since' in self.params or 'until' in self.params

    def get(self):
        # by default, Facebook returns three days 
        # worth' of insights data
        if self._has_daterange:
            seconds = (self.meta['until'] - self.meta['since']).total_seconds()
            days = math.ceil(seconds / 60 / 60 / 24)
        else:
            days = 3

        # TODO: for large date ranges, chunk them up and request
        # the subranges in a batch request (this is a little bit
        # more complex because multiple metrics are *also* batched
        # and so these two things have to work together)
        # 
        # TODO: investigate whether this applies too when asking for 
        # weekly or monthly metrics (that is, whether the limit is 93
        # result rows or truly 93 days)
        if days > 31 * 3:
            raise NotImplementedError(
                "Can only fetch date ranges smaller than 3 months.")

        if 'metrics' in self.meta:
            metrics = []
            for metric in self.meta['metrics']:
                metrics.append({'relative_url': metric})
            results = self.graph.all('insights', 
                metrics, **self.params)
        else:
            results = [self.graph.get('insights', 
                **self.params)]

        datasets = []
        for result in results:
            datasets.extend(result['data'])

        fields = ['end_time']
        data = {}
        for dataset in datasets:
            metric = dataset['name']
            rows = dataset['values']
            fields.append(metric)

            for row in rows:
                value = row['value']
                end_time = row.get('end_time')
                if end_time:
                    end_time = utils.date.parse(end_time)
                else:
                    end_time = 'lifetime'

                data.setdefault(end_time, {})[metric] = value

        Row = namedtuple('Row', fields)

        rows = []
        for time, values in data.items():
            rows.append(Row(end_time=time, **values))

        # when a single metric is requested (and not 
        # wrapped in a list), we return a simplified 
        # data format
        if self.meta['single']:
            metric = self.meta['metrics'][0]
            return [getattr(row, metric) for row in rows]
        else:
            return rows

    def serialize(self):
        return [row._asdict() for row in self.get()]

    def __repr__(self):
        if 'metrics' in self.meta:
            metrics = ", ".join(self.meta['metrics'])
        else:
            metrics = 'all available metrics'

        if self._has_daterange:
            date = ' from {} to {}'.format(
                self.meta['since'].date().isoformat(), 
                self.meta['until'].date().isoformat(), 
                )
        else:
            date = ''

        return u"<Insights for '{}' ({}{})>".format(
            repr(self.edge.name), metrics, date)
        


class Picture(object):
    def __init__(self, post, raw):
        self.post = post
        self.graph = post.graph
        self.raw = self.url = raw
        self.parsed_url = urlparse.urlparse(self.raw)
        self.qs = urlparse.parse_qs(self.parsed_url.query)
        
        if 'url' in self.qs:
            self.origin = self.qs['url'][0]
            self.width = self.qs['w'][0]
            self.height = self.qs['h'][0]
        else:
            self.origin = self.url

        self.basename = self.origin.split('/')[-1]

    def __repr__(self):
        return u"<Picture: {} ({}x{})>".format(
            self.basename, self.width, self.height)


class Post(object):
    def __init__(self, account, raw):
        self.account = account
        self.raw = raw
        # most fields aside from id, type, ctime 
        # and mtime are optional
        self.graph = account.graph.partial(raw['id'])
        self.id = raw['id']
        self.type = raw['type']
        self.created_time = utils.date.parse(raw['created_time'])
        self.updated_time = utils.date.parse(raw['updated_time'])
        self.name = raw.get('name')
        self.story = raw.get('story')
        self.link = raw.get('link')
        self.description = raw.get('description')
        self.shares = raw.get('shares')
        # TODO: figure out if *all* comments and likes are included 
        # when getting post data, or just some
        self.comments = utils.api.getdata(raw, 'comments')
        self.likes = utils.api.getdata(raw, 'likes')
        self.quotes = utils.extract_quotes(self.description or '')
        if 'picture' in raw:
            self.picture = Picture(self, raw['picture'])
        else:
            self.picture = None

    @property
    def insights(self):
        return InsightsSelection(self)

    def resolve_link(self, clean=False):
        if not self.link:
            return None

        url = utils.url.resolve(self.link)

        if clean:
            url = utils.url.base(url)

        return url

    def __repr__(self):
        time = self.created_time.date().isoformat()
        return u"<Post: {} ({})>".format(self.id, time)


class Page(object):
    def __init__(self, token):
        self.graph = GraphAPI(token).partial('me')
        data = self.graph.get()
        self.raw = data
        self.id = data['id']
        self.name = data['name']

    @property
    def token(self):
        return self.graph.oauth_token

    @property
    def insights(self):
        return InsightsSelection(self)

    @property
    def posts(self):
        return PostSelection(self)

    def __repr__(self):
        return u"<Page {}: {}>".format(self.id, self.name)



from setuptools import setup, find_packages

setup(name='facebookinsights',
    description='A wrapper and command-line interface for the Facebook Insights API.',
    #long_description=open('README.rst').read(),
    author='Stijn Debrouwere',
    author_email='stijn@debrouwere.org',
    #url='http://stdbrouw.github.com/facebook-insights/',
    download_url='http://www.github.com/debrouwere/facebook-insights/tarball/master',
    version='0.1.1',
    license='ISC',
    packages=find_packages(),
    keywords='data analytics api wrapper facebook insights',
    entry_points = {
          'console_scripts': [
                'insights = insights.commands:main', 
          ],
    }, 
    install_requires=[
        'click', 
        'requests', 
        'rauth', 
        'facepy', 
        'python-dateutil', 
        'pytz', 
        'addressable', 
        'flask', 
        'keyring', 
    ], 
    # test_suite='facebookinsights.tests', 
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering :: Information Analysis',
        ],
    )